@@ -1,32 +1,36 @@
-from abc import ABC
+from __future__ import annotations
 from typing import Optional, Any, List, Union, Tuple, \
   NamedTuple, Callable, Set
 from pathlib import Path
 from mimetypes import guess_type
+from functools import lru_cache
 
 from pychromecast.controllers.receiver import CastStatus
-from pychromecast.controllers.media import MediaStatus, BaseController
+from pychromecast.controllers.media import MediaStatus, \
+  BaseController, MediaController
 from pychromecast.controllers.youtube import YouTubeController
 from pychromecast.controllers.spotify import SpotifyController
 from pychromecast.controllers.dashcast import DashCastController
+from pychromecast.controllers.bbciplayer import BbcIplayerController
+from pychromecast.controllers.bbcsounds import BbcSoundsController
 # from pychromecast.controllers.homeassistant import HomeAssistantController
 # from pychromecast.controllers.plex import PlexApiController, PlexController
 from pychromecast import Chromecast
 
 from mpris_server.adapters import Metadata, PlayState, \
   Microseconds, VolumeDecimal, RateDecimal
-from mpris_server.base import BEGINNING, DEFAULT_RATE, DbusObj, \
-  Track, Album, Artist
+from mpris_server.base import BEGINNING, DEFAULT_RATE, DbusObj
 from mpris_server.compat import get_dbus_name, enforce_dbus_length
 
+from .types import Protocol, runtime_checkable
 from .base import DEFAULT_THUMB, LIGHT_THUMB, NO_DURATION, NO_DELTA, \
-  US_IN_SEC, DEFAULT_DISC_NO, ChromecastMediaType, DESKTOP_FILE_DARK, \
-  DESKTOP_FILE_LIGHT, NO_DESKTOP_FILE, NAME
+  US_IN_SEC, DEFAULT_DISC_NO, MediaType, NO_DESKTOP_FILE, \
+  NAME, create_desktop_file, DEFAULT_ICON
 
 
-DEFAULT_NAME: str = NAME
 NO_ARTIST: str = ''
 TITLE_SEP: str = ' - '
+MAX_TITLES: int = 3
 
 YOUTUBE_URLS: Set[str] = {
   'youtube.com/',
@@ -37,12 +41,31 @@ YT_VID_URL: str = f'https://{YT_LONG}watch?v='
 
 RESOLUTION: int = 1
 NO_SUFFIX: str = ''
-# NO_DURATION = None
 
 
-class Wrapper(ABC):
-  cc: Chromecast
-  light_icon: bool = False
+class Titles(NamedTuple):
+  title: Optional[str] = None
+  artist: Optional[str] = None
+  album: Optional[str] = None
+
+
+class Controllers(NamedTuple):
+  yt: YouTubeController
+  spotify: SpotifyController
+  dash: DashCastController
+  bbc_ip: BbcIplayerController
+  bbc_sound: BbcSoundsController
+
+
+@runtime_checkable
+class Wrapper(Protocol):
+  dev: Chromecast
+  ctls: Controllers
+  light_icon: bool = DEFAULT_ICON
+
+  @property
+  def name(self) -> str:
+    return self.dev.name or NAME
 
   @property
   def cast_status(self) -> Optional[CastStatus]:
@@ -52,73 +75,72 @@ class Wrapper(ABC):
   def media_status(self) -> Optional[MediaStatus]:
     pass
 
-  def can_play_next(self) -> Optional[bool]:
+  @property
+  def media_controller(self) -> MediaController:
     pass
 
-  def can_play_prev(self) -> Optional[bool]:
-    pass
-
-  def play_next(self):
-    pass
-
-  def play_prev(self):
-    pass
-
-  def open_uri(self, uri: str):
-    pass
-
-  def add_track(
-    self,
-    uri: str,
-    after_track: DbusObj,
-    set_as_current: bool
-  ):
+  @property
+  def titles(self) -> Titles:
     pass
 
 
 class StatusMixin(Wrapper):
   def __getattr__(self, name: str) -> Any:
-    return getattr(self.cc, name)
+    return getattr(self.dev, name)
 
   @property
   def cast_status(self) -> Optional[CastStatus]:
-    if self.cc.status:
-      return self.cc.status
+    if self.dev.status:
+      return self.dev.status
 
     return None
 
   @property
   def media_status(self) -> Optional[MediaStatus]:
-    if self.cc.media_controller.status:
-      return self.cc.media_controller.status
+    if self.media_controller.status:
+      return self.media_controller.status
 
     return None
 
+  @property
+  def media_controller(self) -> MediaController:
+    return self.dev.media_controller
+
 
 class ControllersMixin(Wrapper):
   def __init__(self):
-    self.yt_ctl, self.spotify_ctl, self.dash_ctl = ctls = [
+    self._setup_controllers()
+    super().__init__()
+
+  def _setup_controllers(self):
+    self.ctls = Controllers(
       YouTubeController(),
       SpotifyController(),
       DashCastController(),
-    ]
+      BbcIplayerController(),
+      BbcSoundsController(),
+    )
 
-    for ctl in ctls:
+    for ctl in self.ctls:
       self._register(ctl)
 
-    super().__init__()
-
   def _register(self, controller: BaseController):
-    self.cc.register_handler(controller)
+    self.dev.register_handler(controller)
 
   def _launch_youtube(self):
-    self.yt_ctl.launch()
+    self.ctls.yt.launch()
 
   def _play_youtube(self, video_id: str):
-    if not self.yt_ctl.is_active:
+    if not self.ctls.yt.is_active:
       self._launch_youtube()
 
-    self.yt_ctl.play_video(video_id)
+    self.ctls.yt.play_video(video_id)
+
+  def _is_youtube_vid(self, content_id: str) -> bool:
+    if not content_id or not self.ctls.yt.is_active:
+      return False
+
+    return not content_id.startswith('http')
 
   def _get_url(self) -> Optional[str]:
     content_id = None
@@ -126,7 +148,7 @@ class ControllersMixin(Wrapper):
     if self.media_status:
       content_id = self.media_status.content_id
 
-    if content_id and 'http' not in content_id and self.yt_ctl.is_active:
+    if self._is_youtube_vid(content_id):
       return f'{YT_VID_URL}{content_id}'
 
     return content_id
@@ -139,7 +161,7 @@ class ControllersMixin(Wrapper):
       return
 
     mimetype, _ = guess_type(uri)
-    self.cc.media_controller.play_media(uri, mimetype)
+    self.media_controller.play_media(uri, mimetype)
 
   def add_track(
     self,
@@ -150,29 +172,49 @@ class ControllersMixin(Wrapper):
     video_id = get_video_id(uri)
 
     if video_id:
-      self.yt_ctl.add_to_queue(video_id)
+      self.ctls.yt.add_to_queue(video_id)
 
     if video_id and set_as_current:
-      self.yt_ctl.play_video(video_id)
+      self.ctls.yt.play_video(video_id)
 
     elif set_as_current:
       self.open_uri(uri)
 
 
 class TitlesMixin(Wrapper):
-  def get_titles(self):
-    title = self.cc.media_controller.title
-    subtitle = self.get_subtitle()
-    album = self.media_status.album_name
-    artist = self.media_status.artist
-    app_name = self.cc.app_display_name
+  @property
+  def titles(self) -> Titles:
+    titles: List[str] = list()
+
+    title = self.media_controller.title
+
+    if title:
+      titles.append(title)
 
-  def get_stream_title(self) -> Optional[str]:
-    title = self.cc.media_controller.title
-    app_name = self.cc.app_display_name
     subtitle = self.get_subtitle()
 
-    return title or app_name or subtitle
+    if subtitle:
+      titles.append(subtitle)
+
+    if self.media_status:
+      artist = self.media_status.artist
+
+      if artist:
+        titles.append(artist)
+
+      album = self.media_status.album_name
+
+      if album:
+        titles.append(album)
+
+    app_name = self.dev.app_display_name
+
+    if app_name:
+      titles.append(app_name)
+
+    titles = titles[:MAX_TITLES]
+
+    return Titles(*titles)
 
   def get_subtitle(self) -> Optional[str]:
     if not self.media_status:
@@ -185,60 +227,6 @@ class TitlesMixin(Wrapper):
 
     return None
 
-  def get_artist(self, title: Optional[str] = None) -> Optional[str]:
-    if not title:
-      title = self.get_stream_title()
-
-    subtitle = self.get_subtitle()
-    artist: Optional[str] = None
-
-    if not self.media_status:
-      artist = self.media_status.artist
-
-    app_name: Optional[str] = self.cc.app_display_name
-
-    if artist:
-      return artist
-
-    elif subtitle:
-      return subtitle
-
-    elif app_name and app_name != title:
-      return app_name
-
-    return None
-
-  def get_album(
-    self,
-    title: Optional[str] = None,
-    artist: Optional[str] = None,
-  ) -> Optional[str]:
-    if not title:
-      title = self.get_stream_title()
-
-    if not artist:
-      artist = self.get_artist()
-
-    album: Optional[str] = None
-
-    if not self.media_status:
-      album = self.media_status.album_name
-
-    app_name = self.cc.app_display_name
-    subtitle = self.get_subtitle()
-    titles = {artist, title}
-
-    if album:
-      return album
-
-    elif subtitle and subtitle not in titles:
-      return subtitle
-
-    elif app_name and app_name not in titles:
-      return app_name
-
-    return None
-
 
 class TimeMixin(Wrapper):
   def __init__(self,):
@@ -275,7 +263,8 @@ class TimeMixin(Wrapper):
     position_secs = status.adjusted_current_time
 
     if position_secs:
-      return int(position_secs * US_IN_SEC)
+      position_us = position_secs * US_IN_SEC
+      return round(position_us)
 
     return BEGINNING
 
@@ -290,13 +279,13 @@ class TimeMixin(Wrapper):
     if not status or not status.current_time:
       return False
 
-    rounded = round(status.current_time, RESOLUTION)
+    current_time = round(status.current_time, RESOLUTION)
 
-    return rounded > BEGINNING
+    return current_time > BEGINNING
 
   def seek(self, time: Microseconds):
     seconds = int(round(time / US_IN_SEC))
-    self.cc.media_controller.seek(seconds)
+    self.media_controller.seek(seconds)
 
   def get_rate(self) -> RateDecimal:
     if not self.media_status:
@@ -336,9 +325,9 @@ class IconsMixin(Wrapper):
 
     return str(DEFAULT_THUMB)
 
+  @lru_cache
   def get_desktop_entry(self) -> str:
-    path = \
-      DESKTOP_FILE_LIGHT if self.light_icon else DESKTOP_FILE_DARK
+    path = create_desktop_file(self.light_icon)
 
     if not path:
       return NO_DESKTOP_FILE
@@ -351,8 +340,8 @@ class IconsMixin(Wrapper):
 
 class MetadataMixin(Wrapper):
   def metadata(self) -> Metadata:
-    title: Optional[str] = self.get_stream_title()
-    artist = self.get_artist()
+    title, artist, album = self.titles
+
     artists = [artist] if artist else []
     dbus_name: DbusObj = get_track_id(title)
     comments: List[str] = []
@@ -368,7 +357,7 @@ class MetadataMixin(Wrapper):
       'xesam:url': self._get_url(),
       'xesam:title': title,
       'xesam:artist': artists,
-      'xesam:album': self.get_album(title, artist),
+      'xesam:album': album,
       'xesam:albumArtist': artists,
       'xesam:discNumber': DEFAULT_DISC_NO,
       'xesam:trackNumber': track_no,
@@ -377,48 +366,13 @@ class MetadataMixin(Wrapper):
 
     return metadata
 
-  # TODO: no need to keep this method around
-  # as long as metadata() is implemented
-  def get_current_track(self) -> Track:
-    title = self.get_stream_title()
-    artist_name = self.get_artist()
-    artist = Artist(artist_name)
-    art_url = self.get_art_url()
-    content_id = self._get_url()
-    duration = int(self.get_duration())
-    track_no: Optional[int] = None
-
-    if self.media_status:
-      track_no = self.media_status.track
-
-    album = Album(
-      name=self.get_album(title, artist_name),
-      artists=(artist,),
-      art_url=art_url,
-    )
-
-    track = Track(
-      track_id=get_track_id(title),
-      name=title,
-      track_no=track_no,
-      length=duration,
-      uri=content_id,
-      artists=(artist,),
-      album=album,
-      art_url=art_url,
-      disc_no=DEFAULT_DISC_NO,
-      type=get_media_type(self)
-    )
-
-    return track
-
 
 class PlaybackMixin(Wrapper):
   def get_playstate(self) -> PlayState:
-    if self.cc.media_controller.is_paused:
+    if self.media_controller.is_paused:
       return PlayState.PAUSED
 
-    elif self.cc.media_controller.is_playing:
+    elif self.media_controller.is_playing:
       return PlayState.PLAYING
 
     return PlayState.STOPPED
@@ -436,13 +390,13 @@ class PlaybackMixin(Wrapper):
     return False
 
   def play_next(self):
-    self.cc.media_controller.queue_next()
+    self.media_controller.queue_next()
 
   def play_prev(self):
-    self.cc.media_controller.queue_prev()
+    self.media_controller.queue_prev()
 
   def quit(self):
-    self.cc.quit_app()
+    self.dev.quit_app()
 
   def next(self):
     self.play_next()
@@ -451,16 +405,16 @@ class PlaybackMixin(Wrapper):
     self.play_prev()
 
   def pause(self):
-    self.cc.media_controller.pause()
+    self.media_controller.pause()
 
   def resume(self):
     self.play()
 
   def stop(self):
-    self.cc.media_controller.stop()
+    self.media_controller.stop()
 
   def play(self):
-    self.cc.media_controller.play()
+    self.media_controller.play()
 
   def set_repeating(self, val: bool):
     pass
@@ -470,7 +424,7 @@ class PlaybackMixin(Wrapper):
 
 
 class VolumeMixin(Wrapper):
-  def get_volume(self) -> VolumeDecimal:
+  def get_volume(self) -> Optional[VolumeDecimal]:
     if not self.cast_status:
       return None
 
@@ -478,14 +432,18 @@ class VolumeMixin(Wrapper):
 
   def set_volume(self, val: VolumeDecimal):
     curr = self.get_volume()
-    diff = val - curr
+
+    if curr is None:
+      return
+
+    delta: float = val - curr
 
     # can't adjust vol by 0
-    if diff > NO_DELTA:  # vol up
-      self.cc.volume_up(diff)
+    if delta > NO_DELTA:  # vol up
+      self.dev.volume_up(delta)
 
-    elif diff < NO_DELTA:
-      self.cc.volume_down(abs(diff))
+    elif delta < NO_DELTA:
+      self.dev.volume_down(abs(delta))
 
   def is_mute(self) -> Optional[bool]:
     if self.cast_status:
@@ -494,7 +452,7 @@ class VolumeMixin(Wrapper):
     return False
 
   def set_mute(self, val: bool):
-    self.cc.set_volume_muted(val)
+    self.dev.set_volume_muted(val)
 
 
 class AbilitiesMixin(Wrapper):
@@ -504,7 +462,7 @@ class AbilitiesMixin(Wrapper):
   def can_play(self) -> bool:
     state = self.get_playstate()
 
-    if state is PlayState.PAUSED:
+    if state is not PlayState.PLAYING:
       return True
 
     return False
@@ -543,7 +501,7 @@ class AbilitiesMixin(Wrapper):
     return False
 
 
-class ChromecastWrapper(
+class DeviceWrapper(
   StatusMixin,
   TitlesMixin,
   ControllersMixin,
@@ -554,25 +512,17 @@ class ChromecastWrapper(
   VolumeMixin,
   AbilitiesMixin,
 ):
-  '''
-  A wrapper to make it easier to switch out backend implementations.
+  '''Wraps implementation details for device API'''
 
-  Holds common logic for dealing with underlying Chromecast API.
-  '''
-
-  def __init__(self, cc: Chromecast):
-    self.cc = cc
+  def __init__(self, dev: Chromecast):
+    self.dev = dev
     super().__init__()
 
   def __repr__(self) -> str:
     cls = type(self)
     cls_name = cls.__name__
 
-    return f'<{cls_name} for {self.cc}>'
-
-  @property
-  def name(self) -> str:
-    return self.cc.name or DEFAULT_NAME
+    return f'<{cls_name} for {self.dev}>'
 
 
 @enforce_dbus_length
@@ -581,33 +531,33 @@ def get_track_id(name: str) -> DbusObj:
 
 
 def get_media_type(
-  cc: ChromecastWrapper
-) -> Optional[ChromecastMediaType]:
-  status = cc.media_status
+  dev: DeviceWrapper
+) -> Optional[MediaType]:
+  status = dev.media_status
 
   if not status:
     return None
 
   if status.media_is_movie:
-    return ChromecastMediaType.MOVIE
+    return MediaType.MOVIE
 
   elif status.media_is_tvshow:
-    return ChromecastMediaType.TVSHOW
+    return MediaType.TVSHOW
 
   elif status.media_is_photo:
-    return ChromecastMediaType.PHOTO
+    return MediaType.PHOTO
 
   elif status.media_is_musictrack:
-    return ChromecastMediaType.MUSICTRACK
+    return MediaType.MUSICTRACK
 
   elif status.media_is_generic:
-    return ChromecastMediaType.GENERIC
+    return MediaType.GENERIC
 
   return None
 
 
 def is_youtube(uri: str) -> bool:
-  uri = uri.lower()
+  uri = uri.casefold()
   return any(yt in uri for yt in YOUTUBE_URLS)
 
 
